/*-
 * See the file LICENSE for redistribution information.
 *
 * Copyright (c) 1996, 1997
 *	Sleepycat Software.  All rights reserved.
 */

#include "config.h"

#ifndef lint
static const char sccsid[] = "@(#)bt_stat.c	10.7 (Sleepycat) 7/20/97";
#endif /* not lint */

#ifndef NO_SYSTEM_INCLUDES
#include <sys/types.h>

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#endif

#include "db_int.h"
#include "db_page.h"
#include "btree.h"

/*
 * __bam_stat --
 *	Gather/print the btree statistics
 *
 * PUBLIC: int __bam_stat __P((DB *, void *, void *, void *(*)(size_t)));
 */
int
__bam_stat(argdbp, gspp, lspp, db_malloc)
	DB *argdbp;
	void *gspp, *lspp;
	void *(*db_malloc) __P((size_t));
{
	BTMETA *meta;
	BTREE *t;
	DB *dbp;
	DB_BTREE_LSTAT *lsp;
	DB_BTREE_STAT *gsp;
	DB_LOCK lock;
	PAGE *h;
	db_pgno_t pgno;
	int ret;

	GETHANDLE(argdbp, NULL, &dbp, ret);
	t = dbp->internal;

	/* Return the local statistics. */
	if (lspp != NULL) {
		if ((lsp = db_malloc == NULL ?
		    (DB_BTREE_LSTAT *)malloc(sizeof(*lsp)) :
		    (DB_BTREE_LSTAT *)db_malloc(sizeof(*lsp))) == NULL) {
			ret = ENOMEM;
			goto err;
		}
		*lsp = t->lstat;
		*(DB_BTREE_LSTAT **)lspp = lsp;
	}

	/* The global statistics require a lot more work. */
	if (gspp != NULL) {
		if ((gsp = db_malloc == NULL ?
		    (DB_BTREE_STAT *)malloc(sizeof(*gsp)) :
		    (DB_BTREE_STAT *)db_malloc(sizeof(*gsp))) == NULL) {
			ret = ENOMEM;
			goto err;
		}

		/* Get the meta-data page. */
		pgno = PGNO_METADATA;
		if ((ret = __bam_lget(dbp, 0, pgno, DB_LOCK_READ, &lock)) != 0)
			goto err;
		if ((ret = __bam_pget(dbp, (PAGE **)&meta, &pgno, 0)) != 0)
			goto err;

		/* Start with the meta-data page information. */
		*gsp = meta->stat;

		/* Get the page size from the DB. */
		gsp->bt_pagesize = dbp->pgsize;

		/* Walk the free list, counting pages. */
		for (gsp->bt_free = 0,
		    pgno = meta->free; pgno != PGNO_INVALID;) {
			++gsp->bt_free;

			if ((ret = __bam_pget(dbp, &h, &pgno, 0)) != 0) {
				(void)memp_fput(dbp->mpf, meta, 0);
				(void)__BT_TLPUT(dbp, lock);
				goto err;
			}
			pgno = h->next_pgno;
			(void)memp_fput(dbp->mpf, h, 0);
		}

		/* Discard the meta-data page. */
		(void)memp_fput(dbp->mpf, meta, 0);
		(void)__BT_TLPUT(dbp, lock);

		/* Get the root page. */
		pgno = PGNO_ROOT;
		if ((ret = __bam_lget(dbp,
		    0, PGNO_ROOT, DB_LOCK_READ, &lock)) != 0)
			goto err;
		if ((ret = __bam_pget(dbp, &h, &pgno, 0)) != 0) {
			(void)__BT_LPUT(dbp, lock);
			goto err;
		}

		/* Get the levels from the root page. */
		gsp->bt_levels = h->level;

		/* Walk the database, counting things. */
		for (;;) {
			switch (TYPE(h)) {
			case P_INVALID:
				break;
			case P_IBTREE:
			case P_IRECNO:
				++gsp->bt_int_pg;
				gsp->bt_int_pgfree += HOFFSET(h) - LOFFSET(h);
				break;
			case P_LBTREE:
				++gsp->bt_leaf_pg;
				gsp->bt_leaf_pgfree += HOFFSET(h) - LOFFSET(h);
				gsp->bt_nrecs += NUM_ENT(h) / P_INDX;
				break;
			case P_LRECNO:
				++gsp->bt_leaf_pg;
				gsp->bt_leaf_pgfree += HOFFSET(h) - LOFFSET(h);
				gsp->bt_nrecs += NUM_ENT(h);
				break;
			case P_DUPLICATE:
				++gsp->bt_dup_pg;
				/* XXX gsp->bt_dup_pgfree; */
				break;
			case P_OVERFLOW:
				++gsp->bt_over_pg;
				/* XXX gsp->bt_over_pgfree; */
				break;
			default:
				(void)memp_fput(dbp->mpf, h, 0);
				(void)__BT_LPUT(dbp, lock);
				return (__db_pgfmt(dbp, pgno));
			}

			(void)memp_fput(dbp->mpf, h, 0);
			(void)__BT_LPUT(dbp, lock);

			++pgno;
			if (__bam_lget(dbp, 0, pgno, DB_LOCK_READ, &lock))
				break;
			if (memp_fget(dbp->mpf, &pgno, 0, &h) != 0) {
				(void)__BT_LPUT(dbp, lock);
				break;
			}
		}

		/* Add in the local information. */
		__bam_add_gstat(&t->lstat, gsp);

		*(DB_BTREE_STAT **)gspp = gsp;
	}
	ret = 0;

err:	PUTHANDLE(dbp);
	return (ret);
}

/*
 * __bam_add_gstat --
 *	Add the local statistics to the global statistics.
 *
 * PUBLIC: void __bam_add_gstat __P((DB_BTREE_LSTAT *, DB_BTREE_STAT *));
 */
void
__bam_add_gstat(from, to)
	DB_BTREE_LSTAT *from;
	DB_BTREE_STAT *to;
{
	to->bt_freed += from->bt_freed;
	to->bt_pfxsaved += from->bt_pfxsaved;
	to->bt_split += from->bt_split;
	to->bt_rootsplit += from->bt_rootsplit;
	to->bt_fastsplit += from->bt_fastsplit;
	to->bt_added += from->bt_added;
	to->bt_deleted += from->bt_deleted;
	to->bt_get += from->bt_get;
	to->bt_cache_hit += from->bt_cache_hit;
	to->bt_cache_miss += from->bt_cache_miss;
}
