/*-
 * See the file LICENSE for redistribution information.
 *
 * Copyright (c) 1997
 *	Sleepycat Software.  All rights reserved.
 */

#include "config.h"

#ifndef lint
static const char sccsid[] = "@(#)cxx_table.cpp	10.4 (Sleepycat) 7/20/97";
#endif /* not lint */

#include "db_cxx.h"
#include "cxx_int.h"
#include <errno.h>
#include <string.h>

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                            DBTable                                 //
//                                                                    //
////////////////////////////////////////////////////////////////////////

DBTable::DBTable()
:   imp_(0)
{
}

DBTable::DBTable(const DBTable &that)
:   imp_(that.imp_)
{
}

DBTable &DBTable::operator = (const DBTable &that)
{
    imp_ = that.imp_;
    return *this;
}

DBTable::~DBTable()
{
    // Note: could do automatic close, but only if we reference count
}

int DBTable::isValid() const
{
    return (unwrapConst(this) != 0);
}

int DBTable::kind() const
{
    return DB_UNKNOWN;
}

int DBTable::getFileDescriptor(int *returned_fd)
{
    DB *db = unwrap(this);
    if (!db) {
        DB_ERROR("DBTable::getFileDescriptor", EINVAL);
        return EINVAL;
    }
    int err;
    if ((err = db->fd(db, returned_fd)) != 0) {
        DB_ERROR("DBTable::getFileDescriptor", err);
        return err;
    }
    return 0;
}

int DBTable::sync()
{
    DB *db = unwrap(this);
    if (!db) {
        DB_ERROR("DBTable::sync", EINVAL);
        return EINVAL;
    }
    int err;
    int flags = 0;
    if ((err = db->sync(db, flags)) != 0) {
        DB_ERROR("DBTable::sync", err);
        return err;
    }
    return 0;
}

int DBTable::stat(void *gsp, void *lsp, void *(*alternate_malloc)(size_t))
{
    DB *db = unwrap(this);
    if (!db) {
        DB_ERROR("DBTable::stat", EINVAL);
        return EINVAL;
    }
    int err;
    if ((err = db->stat(db, gsp, lsp, alternate_malloc)) != 0) {
        DB_ERROR("DBTable::stat", err);
        return err;
    }
    return 0;
}

int DBTable::close(int noSync)
{
    DB *db = unwrap(this);
    if (!db) {
        DB_ERROR("DBTable::close", EINVAL);
        return EINVAL;
    }
    int err;
    if ((err = db->close(db, noSync)) != 0) {
        DB_ERROR("DBTable::close", err);
        return err;
    }
    return 0;
}

int DBTable::get(DBThunk &key, DBThunk &value, DBTransaction &txn)
{

    DB *db = unwrap(this);
    int err;
    int flags = 0;

    key.allocate();
    value.allocate();
    if (!db) {
        DB_ERROR("DBTable::get", EINVAL);
        return EINVAL;
    }
    if ((err = db->get(db, unwrap(&txn), unwrap(&key), unwrap(&value), flags)) != 0) {
        DB_ERROR("DBTable::get", err);
        return err;
    }
    return 0;
}

int DBTable::put(DBThunk key, DBThunk value, DBTransaction &txn,
                  int noOverwrites)
{
    DB *db = unwrap(this);
    int err;
    int flags = 0;
    if (noOverwrites)
        flags |= DB_NOOVERWRITE;

    if (!db) {
        DB_ERROR("DBTable::put", EINVAL);
        return EINVAL;
    }
    if ((err = db->put(db, unwrap(&txn), unwrap(&key), unwrap(&value), flags)) != 0) {
        DB_ERROR("DBTable::put", err);
        return err;
    }
    return 0;
}

int DBTable::remove(DBThunk key, DBTransaction &txn)
{
    DB *db = unwrap(this);
    int err;
    int flags = 0;

    if (!db) {
        DB_ERROR("DBTable::remove", EINVAL);
        return EINVAL;
    }
    if ((err = db->del(db, unwrap(&txn), unwrap(&key), flags)) != 0) {
        DB_ERROR("DBTable::remove", err);
        return err;
    }
    return 0;
}

int DBTable::getIterator(DBIterator *iter_returned,
                          DBTransaction &txn)
{
    DB *db = unwrap(this);
    int err;

    if (!db) {
        DB_ERROR("DBTable::getIterator", EINVAL);
        return EINVAL;
    }
    DBC *cursor;
    if ((err = db->cursor(db, unwrap(&txn), &cursor)) != 0) {
        DB_ERROR("DBTable::getIterator", err);
        return err;
    }
    iter_returned->imp_ = wrap(cursor);
    return iter_returned->commonIterate(DB_FIRST, "DBTable::getIterator");
}

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                            DBBTreeTable                            //
//                                                                    //
////////////////////////////////////////////////////////////////////////

DBBTreeTable::DBBTreeTable()
{
}

DBBTreeTable::DBBTreeTable(const DBBTreeTable &that)
:   DBTable(that)
{
}

DBBTreeTable &DBBTreeTable::operator = (const DBBTreeTable &that)
{
    DBTable::operator = (that);
    return *this;
}

DBBTreeTable::~DBBTreeTable()
{
}

int DBBTreeTable::kind() const
{
    return DB_BTREE;
}

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                            DBHashTable                             //
//                                                                    //
////////////////////////////////////////////////////////////////////////

DBHashTable::DBHashTable()
{
}

DBHashTable::DBHashTable(const DBHashTable &that)
:   DBTable(that)
{
}

DBHashTable &DBHashTable::operator = (const DBHashTable &that)
{
    DBTable::operator = (that);
    return *this;
}

DBHashTable::~DBHashTable()
{
}

int DBHashTable::kind() const
{
    return DB_HASH;
}

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                            DBRecnoTable                            //
//                                                                    //
////////////////////////////////////////////////////////////////////////

DBRecnoTable::DBRecnoTable()
{
}

DBRecnoTable::DBRecnoTable(const DBRecnoTable &that)
:   DBTable(that)
{
}

DBRecnoTable &DBRecnoTable::operator = (const DBRecnoTable &that)
{
    DBTable::operator = (that);
    return *this;
}

DBRecnoTable::~DBRecnoTable()
{
}

int DBRecnoTable::kind() const
{
    return DB_RECNO;
}

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                            DBIterator                              //
//                                                                    //
////////////////////////////////////////////////////////////////////////

DBIterator::DBIterator()
:   imp_(0)
{
}

DBIterator::~DBIterator()
{
    // Note: could do automatic close, but only if we reference count
}

int DBIterator::getKey(DBThunk *key_returned)
{
    DBC *cursor = unwrap(this);
    if (!cursor || !currentKey_.hasData()) {
        DB_ERROR("DBTable::getKey", EINVAL);
        return EINVAL;
    }
    key_returned->allocate();
    DBT *result = unwrap(key_returned);
    DBT *key = unwrap(&currentKey_);
    *result = *key;
    return 0;
}

int DBIterator::getValue(DBThunk *value_returned)
{
    DBC *cursor = unwrap(this);
    if (!cursor || !currentValue_.hasData()) {
        DB_ERROR("DBTable::getValue", EINVAL);
        return EINVAL;
    }
    value_returned->allocate();
    DBT *result = unwrap(value_returned);
    DBT *value = unwrap(&currentValue_);
    *result = *value;
    return 0;
}

int DBIterator::get(DBThunk *key_returned, DBThunk *value_returned)
{
    DBC *cursor = unwrap(this);
    if (!cursor || !currentKey_.hasData() || !currentValue_.hasData()) {
        DB_ERROR("DBTable::get", EINVAL);
        return EINVAL;
    }
    key_returned->allocate();
    value_returned->allocate();
    DBT *key_result = unwrap(key_returned);
    DBT *key = unwrap(&currentKey_);
    *key_result = *key;
    DBT *value_result = unwrap(value_returned);
    DBT *value = unwrap(&currentValue_);
    *value_result = *value;
    return 0;
}

//
// common code for put* methods
//
int DBIterator::commonPut(DBThunk key, DBThunk value,
                           int direction, const char *name)
{
    DBC *cursor = unwrap(this);
    if (!cursor) {
        DB_ERROR(name, EINVAL);
        return EINVAL;
    }
    int err;
    if ((err = cursor->c_put(cursor, unwrap(&key), unwrap(&value), direction)) != 0) {
        DB_ERROR(name, err);
        return err;
    }
    return 0;
}

int DBIterator::putCurrent(DBThunk value)
{
    // Note:
    // contrary to documentation, the key value is not
    // totally ignored with the DB_CURRENT flag.  At least,
    // it must not be NULL, so we allocate() it.
    //
    DBThunk ignored;
    ignored.allocate();
    return commonPut(ignored, value, DB_CURRENT, "DBIterator::putCurrent");
}

int DBIterator::putBefore(DBThunk key, DBThunk value)
{
    return commonPut(key, value, DB_AFTER, "DBIterator::putBefore");
}

int DBIterator::putAfter(DBThunk key, DBThunk value)
{
    return commonPut(key, value, DB_AFTER, "DBIterator::putAfter");
}

// common code for next(), previous()
//
int DBIterator::commonIterate(int direction, const char *name)
{
    DBC *cursor = unwrap(this);
    if (!cursor) {
        DB_ERROR(name, EINVAL);
        return EINVAL;
    }
    int err;
    currentKey_.allocate();
    currentValue_.allocate();
    DBT *key = unwrap(&currentKey_);
    DBT *value = unwrap(&currentValue_);
    if ((err = cursor->c_get(cursor, key, value, direction)) != 0) {
        if (err == DB_NOTFOUND) {
            //
            // A deallocated key/value indicates that
            // the iteration is at an end.
            //
            currentKey_.deallocate();
            currentValue_.deallocate();
        }
        else {
            DB_ERROR(name, err);
            return err;
        }
    }
    return err;                 // might be DB_NOTFOUND
}

int DBIterator::next()
{
    return commonIterate(DB_NEXT, "DBIterator::next");
}

int DBIterator::previous()
{
    return commonIterate(DB_NEXT, "DBIterator::previous");
}

int DBIterator::setToBeginning()
{
    return commonIterate(DB_FIRST, "DBTable::setToBeginning");
}

int DBIterator::setToEnd()
{
    return commonIterate(DB_LAST, "DBTable::setToEnd");
}


int DBIterator::noMore()
{
    return (!currentKey_.hasData());
}

int DBIterator::close()
{
    DBC *cursor = unwrap(this);
    if (!cursor) {
        DB_ERROR("DBTable::close", EINVAL);
        return EINVAL;
    }
    int err;
    if ((err = cursor->c_close(cursor)) != 0) {
        DB_ERROR("DBTable::close", err);
        return err;
    }
    return 0;
}

int DBIterator::isValid() const
{
    return (unwrapConst(this) != 0);
}

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                            DBOpenParams                            //
//                                                                    //
////////////////////////////////////////////////////////////////////////

DBOpenParams::DBOpenParams()
:   imp_(0)
,   openFlags_(0)
{
    DB_INFO *info = new DB_INFO;
    memset(info, 0, sizeof(DB_INFO));
    imp_ = wrap(info);
}

DBOpenParams::DBOpenParams(const DBOpenParams &that)
:   imp_(0)
,   openFlags_(0)
{
    DB_INFO *info = new DB_INFO;
    *info = *(unwrapConst(&that));
    imp_ = wrap(info);
    openFlags_ = that.openFlags_;
}

DBOpenParams &DBOpenParams::operator = (const DBOpenParams &that)
{
    openFlags_ = that.openFlags_;
    imp_ = that.imp_;
    return *this;
}

DBOpenParams::~DBOpenParams()
{
    DB_INFO *info = unwrap(this);
    if (info) {
        if (info->re_source) {
            delete [] info->re_source;
        }
        delete info;
    }
}

DB_FLAG_METHODS(DBOpenParams, openFlags_, Create, DB_CREATE)
DB_FLAG_METHODS(DBOpenParams, openFlags_, Readonly, DB_RDONLY)
DB_FLAG_METHODS(DBOpenParams, openFlags_, Truncate, DB_TRUNCATE)

DB_ACCESS_METHODS(DBOpenParams, size_t, CacheSize, DB_INFO, db_cachesize)
DB_ACCESS_METHODS(DBOpenParams, size_t, PageSize, DB_INFO, db_pagesize)
DB_ACCESS_METHODS(DBOpenParams, DBOpenParams::AllocatorFunction,
                   AllocatorFunction, DB_INFO, db_malloc)

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                            DBOpenBTreeParams                       //
//                                                                    //
////////////////////////////////////////////////////////////////////////

DBOpenBTreeParams::DBOpenBTreeParams()
{
}

DBOpenBTreeParams::DBOpenBTreeParams(const DBOpenBTreeParams &that)
:   DBOpenParams(that)
{
}

DBOpenBTreeParams &DBOpenBTreeParams::operator = (const DBOpenBTreeParams &that)
{
    DBOpenParams::operator = (that);
    return *this;
}


DBOpenBTreeParams::~DBOpenBTreeParams()
{
}

DB_ACCESS_METHODS(DBOpenBTreeParams, int, MaxKeysPerPage, DB_INFO, bt_maxkey)
DB_ACCESS_METHODS(DBOpenBTreeParams, int, MinKeysPerPage, DB_INFO, bt_minkey)
DB_ACCESS_METHODS(DBOpenBTreeParams, DBOpenParams::ComparisonFunction,
                   ComparisonFunction, DB_INFO, bt_compare)
DB_ACCESS_METHODS(DBOpenBTreeParams, DBOpenParams::PrefixFunction,
                   PrefixFunction, DB_INFO, bt_prefix)

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                            DBOpenHashParams                        //
//                                                                    //
////////////////////////////////////////////////////////////////////////

DBOpenHashParams::DBOpenHashParams()
{
}

DBOpenHashParams::DBOpenHashParams(const DBOpenHashParams &that)
:   DBOpenParams(that)
{
}

DBOpenHashParams &DBOpenHashParams::operator = (const DBOpenHashParams &that)
{
    DBOpenParams::operator = (that);
    return *this;
}


DBOpenHashParams::~DBOpenHashParams()
{
}

DB_ACCESS_METHODS(DBOpenHashParams, unsigned int,
                  FillFactor, DB_INFO, h_ffactor)
DB_ACCESS_METHODS(DBOpenHashParams, unsigned int,
                  NumberOfElements, DB_INFO, h_nelem)
DB_ACCESS_METHODS(DBOpenHashParams, DBOpenParams::HashFunction,
                   HashFunction, DB_INFO, h_hash)

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                            DBOpenRecnoParams                       //
//                                                                    //
////////////////////////////////////////////////////////////////////////

DBOpenRecnoParams::DBOpenRecnoParams()
{
}

DBOpenRecnoParams::DBOpenRecnoParams(const DBOpenRecnoParams &that)
:   DBOpenParams(that)
{
}

DBOpenRecnoParams &DBOpenRecnoParams::operator = (const DBOpenRecnoParams &that)
{
    DBOpenParams::operator = (that);
    return *this;
}


DBOpenRecnoParams::~DBOpenRecnoParams()
{
}

DB_ACCESS_METHODS_WITH_FLAG(DBOpenRecnoParams, int,
                             PaddingByte, DB_INFO, re_pad, DB_PAD)
DB_ACCESS_METHODS_WITH_FLAG(DBOpenRecnoParams, int,
                             DelimitingByte, DB_INFO, re_delim, DB_DELIMITER)
DB_ACCESS_METHODS_WITH_FLAG(DBOpenRecnoParams, u_int32_t,
                             FixedLength, DB_INFO, re_len, DB_FIXEDLEN)

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                            DBOpenUnknownParams                     //
//                                                                    //
////////////////////////////////////////////////////////////////////////

DBOpenUnknownParams::DBOpenUnknownParams()
{
}

DBOpenUnknownParams::DBOpenUnknownParams(const DBOpenUnknownParams &that)
:   DBOpenParams(that)
{
}

DBOpenUnknownParams &DBOpenUnknownParams::operator = (const DBOpenUnknownParams &that)
{
    DBOpenParams::operator = (that);
    return *this;
}


DBOpenUnknownParams::~DBOpenUnknownParams()
{
}

DB_ACCESS_METHODS(DBOpenUnknownParams, int, MaxKeysPerPage, DB_INFO, bt_maxkey)
DB_ACCESS_METHODS(DBOpenUnknownParams, int, MinKeysPerPage, DB_INFO, bt_minkey)
DB_ACCESS_METHODS(DBOpenUnknownParams, DBOpenParams::ComparisonFunction,
                  ComparisonFunction, DB_INFO, bt_compare)
DB_ACCESS_METHODS(DBOpenUnknownParams, DBOpenParams::PrefixFunction,
                  PrefixFunction, DB_INFO, bt_prefix)
DB_ACCESS_METHODS(DBOpenUnknownParams, unsigned int,
                  FillFactor, DB_INFO, h_ffactor)
DB_ACCESS_METHODS(DBOpenUnknownParams, unsigned int,
                  NumberOfElements, DB_INFO, h_nelem)
DB_ACCESS_METHODS(DBOpenUnknownParams, DBOpenParams::HashFunction,
                  HashFunction, DB_INFO, h_hash)
DB_ACCESS_METHODS_WITH_FLAG(DBOpenUnknownParams, int,
                            PaddingByte, DB_INFO, re_pad, DB_PAD)
DB_ACCESS_METHODS_WITH_FLAG(DBOpenUnknownParams, int,
                            DelimitingByte, DB_INFO, re_delim, DB_DELIMITER)
DB_ACCESS_METHODS_WITH_FLAG(DBOpenUnknownParams, u_int32_t,
                            FixedLength, DB_INFO, re_len, DB_FIXEDLEN)

int DBOpenBTreeParams::hasDuplicateKeys() const
{
    const DB_INFO *info = unwrapConst(this);
    assert(info != 0);
    return ((info->flags & DB_DUP) != 0);
}

void DBOpenBTreeParams::allowDuplicateKeys(int onOrOff)
{
    DB_INFO *info = unwrap(this);
    assert(info != 0);
    if (onOrOff) {
        info->flags |= DB_DUP;
    }
    else {
        info->flags &= ~(DB_DUP);
    }
}

int DBOpenHashParams::hasDuplicateKeys() const
{
    const DB_INFO *info = unwrapConst(this);
    assert(info != 0);
    return ((info->flags & DB_DUP) != 0);
}

void DBOpenHashParams::allowDuplicateKeys(int onOrOff)
{
    DB_INFO *info = unwrap(this);
    assert(info != 0);
    if (onOrOff) {
        info->flags |= DB_DUP;
    }
    else {
        info->flags &= ~(DB_DUP);
    }
}

void DBOpenRecnoParams::setSourceFileName(const char *value)
{
    DB_INFO *info = unwrap(this);
    assert(info != 0);
    info->re_source = new char[strlen(value) + 1];
    strcpy(info->re_source, value);
}

const char *DBOpenRecnoParams::getSourceFileName() const
{
    const DB_INFO *info = unwrapConst(this);
    assert(info != 0);
    return info->re_source;
}

int DBOpenRecnoParams::hasSnapshot() const
{
    const DB_INFO *info = unwrapConst(this);
    assert(info != 0);
    return ((info->flags & DB_SNAPSHOT) != 0);
}

void DBOpenRecnoParams::useSnapshot(int onOrOff)
{
    DB_INFO *info = unwrap(this);
    assert(info != 0);
    if (onOrOff) {
        info->flags |= DB_SNAPSHOT;
    }
    else {
        info->flags &= ~(DB_SNAPSHOT);
    }
}

////////////////////////////////////////////////////////////////////////
//                                                                    //
//                            DBThunk                                 //
//                                                                    //
////////////////////////////////////////////////////////////////////////

DBThunk::DBThunk()
:   imp_(0)
{
    // not allocated by default!
}

DBThunk::DBThunk(void *data, size_t size)
:   imp_(0)
{
    setData(data, size);
}

DBThunk::~DBThunk()
{
    deallocate();
}

DBThunk::DBThunk(const DBThunk &that)
:   imp_(0)
{
    const DBT *from = unwrapConst(&that);
    if (from) {
        allocate();
        DBT *to = unwrap(this);
        *to = *from;                // structure assignment
    }
}

DBThunk &DBThunk::operator = (const DBThunk &that)
{
    const DBT *from = unwrapConst(&that);
    if (from) {
        allocate();
        DBT *to = unwrap(this);
        *to = *from;                // structure assignment
    }
    else {
        deallocate();
    }
    return *this;
}

void DBThunk::setData(void *data, size_t size)
{
    allocate();
    DBT *dbt = unwrap(this);
    dbt->data = data;
    dbt->size = size;
}

void DBThunk::setData(void *data)
{
    allocate();
    DBT *dbt = unwrap(this);
    dbt->data = data;
}

void DBThunk::setSize(size_t size)
{
    allocate();
    DBT *dbt = unwrap(this);
    dbt->size = size;
}

void *DBThunk::getData() const
{
    const DBT *dbt = unwrapConst(this);
    if (!dbt)
        return 0;
    return dbt->data;
}

size_t DBThunk::getSize() const
{
    const DBT *dbt = unwrapConst(this);
    if (!dbt)
        return 0;
    return dbt->size;
}

void DBThunk::setUserBufferSize(size_t size)
{
    DBT *dbt = unwrap(this);
    dbt->ulen = size;
}

size_t DBThunk::getUserBufferSize() const
{
    const DBT *dbt = unwrapConst(this);
    if (!dbt)
        return 0;
    return dbt->ulen;
}

size_t DBThunk::getRecordSize() const
{
    const DBT *dbt = unwrapConst(this);
    if (!dbt)
        return 0;
    return dbt->dlen;
}

size_t DBThunk::getRecordOffset() const
{
    const DBT *dbt = unwrapConst(this);
    if (!dbt)
        return 0;
    return dbt->doff;
}

void DBThunk::setReturnInAllocatedMemory(int onOrOff)
{
    allocate();
    DBT *dbt = unwrap(this);
    if (onOrOff) {
        dbt->flags |= DB_DBT_MALLOC;
    }
    else {
        dbt->flags &= ~(DB_DBT_MALLOC);
    }
}

int DBThunk::getReturnInAllocatedMemory() const
{
    const DBT *dbt = unwrapConst(this);
    if (!dbt)
        return 0;
    return (dbt->flags & DB_DBT_MALLOC) ? 1 : 0;
}

void DBThunk::setReturnInUserBuffer(int onOrOff)
{
    allocate();
    DBT *dbt = unwrap(this);
    if (onOrOff) {
        dbt->flags |= DB_DBT_USERMEM;
    }
    else {
        dbt->flags &= ~(DB_DBT_USERMEM);
    }
}

int DBThunk::getReturnInUserBuffer() const
{
    const DBT *dbt = unwrapConst(this);
    if (!dbt)
        return 0;
    return (dbt->flags & DB_DBT_USERMEM) ? 1 : 0;
}

void DBThunk::setPartial(int onOrOff)
{
    allocate();
    DBT *dbt = unwrap(this);
    if (onOrOff) {
        dbt->flags |= DB_DBT_PARTIAL;
    }
    else {
        dbt->flags &= ~(DB_DBT_PARTIAL);
    }
}

int DBThunk::getPartial() const
{
    const DBT *dbt = unwrapConst(this);
    if (!dbt)
        return 0;
    return (dbt->flags & DB_DBT_PARTIAL) ? 1 : 0;
}


void DBThunk::allocate()
{
    if (imp_)
        return;

    DBT *dbt = new DBT;
    memset(dbt, 0, sizeof(DBT));
    imp_ = wrap(dbt);
}

void DBThunk::deallocate()
{
    if (!imp_)
        return;

    DBT *dbt = unwrap(this);
    delete dbt;
    imp_ = 0;
}

int DBThunk::hasData() const
{
    return (imp_ != 0);
}
