/*-
 * See the file LICENSE for redistribution information.
 *
 * Copyright (c) 1997
 *	Sleepycat Software.  All rights reserved.
 *
 *	@(#)db_cxx.h	10.5 (Sleepycat) 7/27/97
 */

#ifndef _DB_CXX_H_
#define _DB_CXX_H_

//
// C++ assumptions:
//
// To ensure portability to many platforms, both new and old, we make
// few assumptions about the C++ compiler and library.  For example,
// we do not expect STL, templates or namespaces to be available.  The
// "newest" C++ feature used is exceptions, which are used liberally
// to transmit error information.  Even the use of exceptions can be
// disabled at runtime, see setErrorModel().
//
// C++ naming conventions:
//
// All top level class names start with DB.
// All non-static class members start with lower case letter.
// All static class members start with upper case letter.
// All private data members are suffixed with underscore.
// Use uppercase letters to divide names into multiple words.
// Accessors are named with get or set prefix.
//
// Object ownership conventions:
//
// Accepting a non-const pointer to an object means taking responsibility
// for that object.
//
//    If an object's member function takes a non-const pointer as a parameter,
//    then the object takes ownership of the memory that the pointer points to.
//
//    If an object calls a function that returns a non-const pointer,
//    then the object takes ownership of the memory that the pointer points to.
//
// Accepting a reference or const pointer means no transfer of responsibility.
//
//
// Generally, the DB classes represent handles to structures that
// are invisible to users of the interface.  Therefore, they can
// usually be copied or passed as arguments, since a shallow copy
// of the handle is done.  This library does no reference counting,
// so it is up to the user to properly call the "close" method when
// finished with a handle.
//
// Exceptions to the above include DBApp which closes the application
// automatically on destruction.  It cannot be copied.
// The classes DB*Options, DB*Params and DBThunk can
// be copied, but the copy is deep (complete), so they
// do not behave as handles.
//
// In cases where a handle is initialized (e.g. first arg of
// DbApp::openHashTable) the user should declare a handle (using the
// default constructor) and pass in its address.  See the examples for
// details.
//

////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
//
// Forward declarations
//

//
// The db C++ interface is almost completely independent of types in C
// portion of db.  Still, there are a few simple types like db_recno_t,
// and larger informational structures like db_mpool_stat that are needed
// from db.h .
//
#include "db.h"

class DBApp;                                     // forward
class DBAppOptions;                              // forward
class DBBTreeTable;                              // forward
class DBConfig;                                  // forward
class DBException;                               // forward
class DBHashTable;                               // forward
class DBIterator;                                // forward
class DBLock;                                    // forward
class DBLockTable;                               // forward
class DBLockRequest;                             // forward
class DBLog;                                     // forward
class DBMemPool;                                 // forward
class DBMemPoolFile;                             // forward
class DBMemPoolFileStat;                         // forward
class DBMemPoolStat;                             // forward
class DBOpenType;                                // forward
class DBOpenParams;                              // forward
class DBOpenBTreeCapabilities;                   // forward
class DBOpenBTreeHashCapabilities;               // forward
class DBOpenBTreeParams;                         // forward
class DBOpenHashCapabilities;                    // forward
class DBOpenHashParams;                          // forward
class DBOpenRecnoCapabilities;                   // forward
class DBOpenRecnoParams;                         // forward
class DBRecnoTable;                              // forward
class DBSequenceNumber;                          // forward
class DBTable;                                   // forward
class DBThunk;                                   // forward
class DBTransaction;                             // forward
class DBTransactionManager;                      // forward
class DBVersion;                                 // forward

class DBAppImp;                                  // forward
class DBAppOptionsImp;                           // forward
class DBIteratorImp;                             // forward
class DBLockRequestImp;                          // forward
class DBLockTableImp;                            // forward
class DBLogImp;                                  // forward
class DBMemPoolImp;                              // forward
class DBMemPoolFileImp;                          // forward
class DBMemPoolFileStatImp;                      // forward
class DBMemPoolStatImp;                          // forward
class DBOpenParamsImp;                           // forward
class DBSequenceNumberImp;                       // forward
class DBTableImp;                                // forward
class DBThunkImp;                                // forward
class DBTransactionImp;                          // forward
class DBTransactionManagerImp;                   // forward

////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
//
// Mechanisms for declaring classes
//

//
// Every class defined in DB has an _exported next to the class name.
// This is needed for WinTel machines so that the class methods can
// be exported or imported in a DLL as appropriate.  Users of the DLL
// use the define DB_USE_DLL.  When the DLL is built, DB_CREATE_DLL
// must be defined.
//
#if defined(_MSC_VER)

#  if defined(DB_CREATE_DLL)
#    define _exported __declspec(dllexport)      // creator of dll
#  elif defined(DB_USE_DLL)
#    define _exported __declspec(dllimport)      // user of dll
#  else
#    define _exported                            // static lib creator or user
#  endif

#else

#  define _exported

#endif

//
// DEFINE_DB_CLASS_WITH_TYPE defines an imp_ data member
// with an "imp()" accessor.
//
// DEFINE_DB_CLASS is the 'typical' case, where the underlying
// type is a pointer to an opaque *Imp class, that gets converted
// to the correct implementation class by the implementation.
//
// Since these defines use "private/public" labels, and leave the access
// being "private", we always use these by convention before any data
// members in the private section of a class.  Keeping them in the
// private section also emphasizes that they are off limits to user code.
//
#define DEFINE_DB_CLASS_WITH_TYPE(name, type) \
    public: type imp() { return imp_; } \
    public: const type imp() const { return imp_; } \
    private: type imp_

#define DEFINE_DB_CLASS(name)   DEFINE_DB_CLASS_WITH_TYPE(name, name##Imp*)


////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
//
// Turn off inappropriate compiler warnings
//

#ifdef _MSC_VER

// These are level 4 warnings that are explicitly disabled.
// With Visual C++, by default you do not see above level 3 unless
// you use /W4.  But we like to compile with the highest level
// warnings to catch other errors.
//
// 4201: nameless struct/union
//       triggered by standard include file <winnt.h>
//
// 4514: unreferenced inline function has been removed
//       certain include files in MSVC define methods that are not called
//
#pragma warning(disable: 4201 4514)

#endif

////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
//
// Exception classes
//

// Almost any error in the DB library throws a DBException.
// Except for the subclass DBKeyExistsException, which may
// arise in normal operation, every error should be considered
// an abnormality (e.g. bug, misuse of DB, file system error).
//
// NOTE: We would like to inherit from class exception and
//       let it handle what(), but there are
//       MSVC++ problems when <exception> is included.
//
class _exported DBException
{
public:
    virtual ~DBException();
    DBException(int err);
    DBException(const char *description);
    DBException(const char *prefix, int err);
    DBException(const char *prefix1, const char *prefix2, int err);
    const int getErrno();
    virtual const char *what() const;

private:
    char *what_;
    int err_;                   // errno
};

class _exported DBKeyExistsException : public DBException
{
public:
    DBKeyExistsException(const char *description);
};


////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
//
// Lock classes
//

class _exported DBLock
{
    friend DBLockTable;
public:
    DBLock(unsigned int);
    DBLock();
    unsigned int getId();
    void setId(unsigned int);

private:
    DEFINE_DB_CLASS_WITH_TYPE(DBLock, unsigned int);
};

class _exported DBLockRequest
{
    friend DBLockTable;
public:

    typedef u_int32_t LockerId;

    typedef int LockMode;
    // one of the following (do not "or"):
    static const int ModeNotGranted;
    static const int ModeSharedRead;
    static const int ModeExclusiveWrite;
    static const int ModeIntendRead;
    static const int ModeIntendWrite;
    static const int ModeIntendReadWrite;

    typedef int LockOp;
    // one of the following (do not "or"):
    static const int OpDump;    // Display held locks.
    static const int OpGet;     // Get the lock.
    static const int OpPut;     // Release the lock.
    static const int OpPutAll;  // Release locker's locks.
    static const int OpPutObj;  // Release locker's locks on obj.

    DBLockRequest(LockOp op,
	LockMode mode, LockerId locker, const DBThunk &obj);
    DBLock lock();             // lock returned

    int isValid() const;

private:
    DEFINE_DB_CLASS(DBLockRequest);
};

class _exported DBLockTable
{
friend DBApp;
public:
    typedef DBLockRequest::LockerId LockerId;
    typedef DBLockRequest::LockMode LockMode;

    // Note: flags for constructor are the same as those used by lock_open()
    //
    DBLockTable(const char *dir, int flags, int mode, DBApp &app);
    ~DBLockTable();

    int close();

    // Note: flags for get are the same as those used by lock_get()
    //
    int get(LockerId locker, int flags, const DBThunk &obj,
            LockMode lock_mode, DBLock *lock_returned);
    int id(LockerId *id_returned);
    int put(DBLock lock);

    // Note: flags for vector are the same as those used by lock_vec()
    //
    int vector(LockerId locker, int flags, DB_LOCKREQ list[],
	    int nlist, DB_LOCKREQ **elist_returned);

    int isValid() const;

private:
    DBLockTable();             // used by DBApp

    DEFINE_DB_CLASS(DBLockTable);
};


////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
//
// Log classes
//

class _exported DBLog
{
friend DBApp;
public:
    DBLog(const char *dir, int flags, int mode, DBApp &dbapp);
    int close();
    static int Compare(DBSequenceNumber sn1, DBSequenceNumber sn2);
    int file(DBSequenceNumber sn, char *name_returned, int name_len);
    int flush(DBSequenceNumber sn);
    int get(DBSequenceNumber sn, DBThunk *data_returned, int flags);
    int put(DBSequenceNumber *sn_returned, const DBThunk &data, int flags);
    int identify(DBTable &dbp, const char *name, u_int32_t *fid_returned);
    int unidentify(u_int32_t fid);

    int isValid() const;

private:
    DBLog();             // used by DBApp

    DEFINE_DB_CLASS(DBLog);
};


////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
//
// Memory pool classes
//

class _exported DBMemPoolFile
{
public:
    // Flags for constructor:
    static const int CreateFile;
    static const int NoMmap;
    static const int ReadOnly;

    DBMemPoolFile(DBMemPool &mp, const char *file,
                   int ftype, int flags, int mode,
                   size_t pagesize, int lsn_offset,
                   DBThunk &pgcookie);

    virtual ~DBMemPoolFile();

    // Flags for get:
    static const int Create;
    static const int Last;
    static const int New;

    // Flags for put/set:
    static const int Clean;
    static const int Dirty;
    static const int Discard;

    int close();
    int get(db_pgno_t *pgnoaddr, unsigned long flags, void *pagep);
    int put(void *pgaddr, unsigned long flags);
    int set(void *pgaddr, unsigned long flags);
    int sync();

private:
    DEFINE_DB_CLASS(DBMemPoolFile);
};

class _exported DBMemPool
{
friend DBApp;
public:
    // Flags for constructor:
    static const int Create;

    DBMemPool(const char *dir, int flags, int mode, DBApp &app);
    virtual ~DBMemPool();

    int close();

    // access to low level interface
    int identify(int ftype,
                 int (*pgin)(db_pgno_t pgno, void *pgaddr, DBT *pgcookie),
                 int (*pgout)(db_pgno_t pgno, void *pgaddr, DBT *pgcookie));

    int stat(DB_MPOOL_STAT **gsp, DB_MPOOL_FSTAT ***fsp,
             void *(*alternate_malloc)(size_t));
    int sync(DBSequenceNumber &);

    int isValid() const;

private:
    DBMemPool();             // used by DBApp

    DEFINE_DB_CLASS(DBMemPool);
};


////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
//
// Transaction classes
//

class _exported DBTransactionManager
{
friend DBApp;
public:
    // Note: flags for constructor are the same as those used by txn_open()
    //
    DBTransactionManager(const char *dir, int flags, int mode, DBApp &dbenv);
    int close();
    int checkpoint(long kbyte, long min) const;

    int isValid() const;

private:
    DBTransactionManager();             // used by DBApp

    DEFINE_DB_CLASS(DBTransactionManager);
};

class _exported DBTransaction
{
public:
    static DBTransaction NoTransaction;

    // Note:
    // Default constructor is only used to create the special
    // NoTransaction object.  Use the constructor with arguments.
    //
    DBTransaction();

    DBTransaction(DBTransactionManager &manager,
                   DBTransaction nestedParent = NoTransaction);
    int abort();
    int commit();
    u_int32_t id();
    int prepare();

    int isValid() const;

private:
    DEFINE_DB_CLASS(DBTransaction);
};


////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
//
// Open parameter classes
//


class _exported DBOpenParams
{
    friend DBApp;

public:
    //
    // create the underlying file as necessary.
    //
    void setCreate(int onOrOff);
    int getCreate() const;

    //
    // set the readonly flag, so that write operations are not permitted.
    // This flag is ignored when used with createTable();
    //
    void setReadonly(int onOrOff);
    int getReadonly() const;

    //
    // set the truncate flag, so that database is truncated on create.
    // This flag is ignored when used with DBApp::open*Table();
    //
    void setTruncate(int onOrOff);
    int getTruncate() const;

    //
    // explicitly set a byte order for the database.
    // By default, a byte order appropriate for the host
    // machine is chosen.
    //
    void setByteOrderToHostMachine();
    void setByteOrderLittleEndian();
    void setByteOrderBigEndian();
    int isByteOrderHostMachine() const;
    int isByteOrderLittleEndian() const;
    int isByteOrderBigEndian() const;

    //
    // suggested cache size for memory pool.
    // arg of zero sets to appropriate default.
    //
    void setCacheSize(size_t);
    size_t getCacheSize() const;

    //
    // suggested page size for database.
    // arg of zero sets to appropriate default.
    //
    void setPageSize(size_t);
    size_t getPageSize() const;

    //
    // the user can provide an alternate "malloc()", or
    // various other functions that db normally provides.
    //
    typedef void *(*AllocatorFunction)(size_t);

    void setAllocatorFunction(AllocatorFunction);
    AllocatorFunction getAllocatorFunction() const;

    // These typedefs are used by subclasses DBOpenBTreeParams
    // and DBOpenUnknownParams
    //
    typedef int (*ComparisonFunction)(const DBT *, const DBT *);
    typedef size_t (*PrefixFunction)(const DBT *, const DBT *);

    // This typedefs are used by subclasses DBOpenHashParams
    // and DBOpenUnknownParams
    //
    typedef u_int32_t (*HashFunction)(const void *, u_int32_t);


    virtual ~DBOpenParams();
    DBOpenParams();
    DBOpenParams(const DBOpenParams &);
    DBOpenParams &operator = (const DBOpenParams &);

private:
    DEFINE_DB_CLASS(DBOpenParams);

protected:
    int openFlags_;                  // open flags, currently only RDONLY
};

class _exported DBOpenBTreeParams : public DBOpenParams
{
public:
    //
    // methods specific to BTree access.
    //
    int getMaxKeysPerPage() const;
    void setMaxKeysPerPage(int value);

    int getMinKeysPerPage() const;
    void setMinKeysPerPage(int value);

    ComparisonFunction getComparisonFunction() const;
    void setComparisonFunction(ComparisonFunction value);

    PrefixFunction getPrefixFunction() const;
    void setPrefixFunction(PrefixFunction value);

    int hasDuplicateKeys() const;
    void allowDuplicateKeys(int onOrOff);

    DBOpenBTreeParams();
    ~DBOpenBTreeParams();
    DBOpenBTreeParams(const DBOpenBTreeParams &);
    DBOpenBTreeParams &operator = (const DBOpenBTreeParams &);
};


class _exported DBOpenHashParams : public DBOpenParams
{
public:
    //
    // methods specific to Hash access.
    //
    unsigned int getFillFactor() const;
    void setFillFactor(unsigned int value);

    unsigned int getNumberOfElements() const;
    void setNumberOfElements(unsigned int value);

    HashFunction getHashFunction() const;
    void setHashFunction(HashFunction value);

    int hasDuplicateKeys() const;
    void allowDuplicateKeys(int onOrOff);

    DBOpenHashParams();
    ~DBOpenHashParams();
    DBOpenHashParams(const DBOpenHashParams &);
    DBOpenHashParams &operator = (const DBOpenHashParams &);
};

class _exported DBOpenRecnoParams : public DBOpenParams
{
public:
    //
    // methods specific to Recno access.
    //
    int hasPaddingByte() const;
    int getPaddingByte() const;
    void setPaddingByte(int value);

    int hasDelimitingByte() const;
    int getDelimitingByte() const;
    void setDelimitingByte(int value);

    int hasFixedLength() const;
    u_int32_t getFixedLength() const;
    void setFixedLength(u_int32_t value);

    const char *getSourceFileName() const;
    void setSourceFileName(const char *value);

    int hasSnapshot() const;
    void useSnapshot(int onOrOff);

    DBOpenRecnoParams();
    ~DBOpenRecnoParams();
    DBOpenRecnoParams(const DBOpenRecnoParams &);
    DBOpenRecnoParams &operator = (const DBOpenRecnoParams &);
};

class _exported DBOpenUnknownParams : public DBOpenParams
{
public:
    //
    // methods specific to BTree access.
    //
    int getMaxKeysPerPage() const;
    void setMaxKeysPerPage(int value);

    int getMinKeysPerPage() const;
    void setMinKeysPerPage(int value);

    ComparisonFunction getComparisonFunction() const;
    void setComparisonFunction(ComparisonFunction value);

    PrefixFunction getPrefixFunction() const;
    void setPrefixFunction(PrefixFunction value);

    //
    // methods specific to Hash access.
    //
    unsigned int getFillFactor() const;
    void setFillFactor(unsigned int value);

    unsigned int getNumberOfElements() const;
    void setNumberOfElements(unsigned int value);

    HashFunction getHashFunction() const;
    void setHashFunction(HashFunction value);

    //
    // methods specific to either BTree or Hash access
    //
    int hasDuplicateKeys() const;
    void allowDuplicateKeys(int onOrOff);

    //
    // methods specific to Recno access.
    //
    int hasPaddingByte() const;
    int getPaddingByte() const;
    void setPaddingByte(int value);

    int hasDelimitingByte() const;
    int getDelimitingByte() const;
    void setDelimitingByte(int value);

    int hasFixedLength() const;
    u_int32_t getFixedLength() const;
    void setFixedLength(u_int32_t value);

    const char *getSourceFileName() const;
    void setSourceFileName(const char *value);

    int hasSnapshot() const;
    void useSnapshot(int onOrOff);

    DBOpenUnknownParams();
    ~DBOpenUnknownParams();
    DBOpenUnknownParams(const DBOpenUnknownParams &);
    DBOpenUnknownParams &operator = (const DBOpenUnknownParams &);
};

////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
//
// Application classes
//

//
// A set of application options - define how this application uses
// the db library.
//
class _exported DBAppOptions
{
    friend DBApp;

public:
    DBAppOptions();
    ~DBAppOptions();

    //
    // Does this application need underlying files that
    // implement Locking, Logging, etc. to be created?
    // If not, then the user or another application
    // has already created them.
    //
    void setCreateFiles(int onOrOff);
    int getCreateFiles() const;

    //
    // Does this application use file locking?
    //
    void setUseLocking(int onOrOff);
    int getUseLocking() const;

    //
    // Does this application use logging to allow for recovery?
    //
    void setUseLogging(int onOrOff);
    int getUseLogging() const;

    //
    // Does this application use memory pools?
    // These are needed for DB access methods.
    //
    void setUseMemPools(int onOrOff);
    int getUseMemPools() const;

    //
    // Does this application use transactions?
    //
    void setUseTransactions(int onOrOff);
    int getUseTransactions() const;

    //
    // Should the process' environment be used
    // to help specify file naming?
    //
    void setUseEnvironment(int onOrOff);
    int getUseEnvironment() const;

    //
    // Should the process' environment be used
    // to help specify file naming when the user is root?
    //
    void setUseEnvironmentForRoot(int onOrOff);
    int getUseEnvironmentForRoot() const;

    // The default error model is to throw an exception whenever
    // an error occurs.  This generally allows for cleaner logic
    // for transaction processing, as a try block can surround a
    // single transaction.  Alternatively, since almost every method
    // returns an error code (errno), the error model can be set to
    // not throw exceptions, and instead return the appropriate code.
    //
    enum ErrorModel { Exception, ErrorReturn };
    void setErrorModel(ErrorModel);
    ErrorModel getErrorModel() const;

    // If an error is detected and the error call function
    // or stream is set, a message is dispatched or printed.
    // If a prefix is set, each message is prefixed.
    //
    // The three ways to set the dispatching of errors have
    // the following precedence:
    //
    //  The last of either setErrorFunction or setErrorStream
    //  that is called sets the error dispatching.  If neither are called,
    //  or is last called with a zero argument, the FILE* error stream is
    //  used.
    //
    // Note that calling setErrorStream has the effect of internally
    // setting the ErrorFunction, and also sets the stream for all DBApp
    // structures.  Note also that setting the error stream does so
    // for all DBApps, although it is anticipated that there is only
    // one DBApp.
    //
    typedef void (*ErrorFunction)(const char *db_errpfx, char *buffer);
    void setErrorFunction(ErrorFunction);
    ErrorFunction getErrorFunction() const;

    class ostream* getErrorStream() const;
    void setErrorStream(class ostream*); // see comment above for precedence

    FILE *getErrorFile() const;
    void setErrorFile(FILE *);              // see comment above for precedence

    void setErrorPrefix(const char *prefix);
    const char *getErrorPrefix() const;

    void setErrorBuffer(char *buffer);

    // Set the yield function, called just before the db library
    // needs to wait on a lock.
    //
    typedef int (*YieldFunction)();
    void setYieldFunction(YieldFunction);
    YieldFunction getYieldFunction() const;

    // Set parameters for locking, if locking is used.
    //
    void setLockingConflictMatrix(u_int8_t *matrix);
    u_int8_t *getLockingConflictMatrix() const;

    void setLockModes(int nLockModes);
    int getLockModes() const;

    void setMaxLocks(unsigned int maxNumberOfLocks);
    unsigned int getMaxLocks() const;


    // Set parameters for logging, if logging is used.
    //
    void setMaxLogFileSize(u_int32_t sizeInBytes);
    u_int32_t getMaxLogFileSize() const;

    // Set parameters for memory pools, if mempools are used.
    //
    void setMaxMemPoolMmapSize(size_t mmapsize);
    size_t getMaxMemPoolMmapSize() const;

    void setMemPoolCacheSize(size_t cachesize);
    size_t getMemPoolCacheSize() const;

    void setPrivateMemPool(int onOrOff);
    int getPrivateMemPool() const;


    // Set parameters for transactions, if transactions are used.
    //
    typedef int (*TransactionRecoveryFunction)
        (DB_LOG *, DBT *, DB_LSN *, int, void *);
    void setTransactionRecoveryFunction(TransactionRecoveryFunction);
    TransactionRecoveryFunction getTransactionRecoveryFunction() const;

    void setMaxTransactions(unsigned int maxNumberOfTransactions);
    unsigned int getMaxTransactions() const;


    // (deep) copying of this object is allowed.
    //
    DBAppOptions(const DBAppOptions &);
    DBAppOptions &operator = (const DBAppOptions &);

private:
    static void StreamErrorFunction(const char *, char *);

    DEFINE_DB_CLASS(DBAppOptions);

    int options_;               // options not covered by dbenv
    ErrorModel errorModel_;
    static ostream *errorStream_;
};

//
// Base application class.  Provides functions for opening a database.
// User of this library can use this class as a starting point for
// developing a DB application - derive their application class from
// this one, add application control logic.
//
class _exported DBApp
{
public:
    static const DBAppOptions DefaultAppOptions;
    static const DBOpenBTreeParams DefaultOpenBTreeParams;
    static const DBOpenHashParams DefaultOpenHashParams;
    static const DBOpenRecnoParams DefaultOpenRecnoParams;
    static const DBOpenUnknownParams DefaultOpenUnknownParams;

    DBApp(const char *homeDir, DBAppOptions options = DefaultAppOptions);
    DBApp(const char *homeDir, DBAppOptions options, const DBConfig &config);
    virtual ~DBApp();

    // If the name is set, it is used to as a prefix (with ": ")
    // before any error messages (see setErrorStream).
    //
    void setName(const char *);

    // Change the error model for handling errors.
    // See also DBAppOptions::setErrorModel().
    //
    void setErrorModel(DBAppOptions::ErrorModel);
    DBAppOptions::ErrorModel getErrorModel() const;

    // int mode  is UNIX-style octal protection flags.
    // Investigate: <stat.h> for POSIX alternative.
    //
    int createBTreeTable(DBBTreeTable *table_returned,
                         const char *fname,
                         int mode,
                         DBOpenBTreeParams params = DefaultOpenBTreeParams);

    int createHashTable(DBHashTable *table_returned,
                        const char *fname,
                        int mode,
                        DBOpenHashParams params = DefaultOpenHashParams);

    int createRecnoTable(DBRecnoTable *table_returned,
                         const char *fname,
                         int mode,
                         DBOpenRecnoParams params = DefaultOpenRecnoParams);

    int openBTreeTable(DBBTreeTable *table_returned,
                       const char *fname,
                       DBOpenBTreeParams params = DefaultOpenBTreeParams);

    int openHashTable(DBHashTable *table_returned,
                      const char *fname,
                      DBOpenHashParams params = DefaultOpenHashParams);

    int openRecnoTable(DBRecnoTable *table_returned,
                       const char *fname,
                       DBOpenRecnoParams params = DefaultOpenRecnoParams);

    int openUnknownTable(DBTable *table_returned,
                         const char *fname,
                         DBOpenUnknownParams params = DefaultOpenUnknownParams);


    // Get the locking table, which will only be valid
    // if the user specified locking in the DBAppOptions.
    //
    DBLockTable getLockTable() const;

    // Get the memory pool, which will only be valid
    // if the user specified memory pools in the DBAppOptions.
    //
    DBMemPool getMemPool() const;

    // Get the transaction manager, which will only be valid
    // if the user specified transactions in the DBAppOptions.
    //
    DBTransactionManager getTransactionManager() const;

    // Get the log, which will only be valid
    // if the user specified logging in the DBAppOptions.
    //
    DBLog getLog() const;

    // Remove the underlying files for the various subsystems.
    //
    int removeLockTableFile(const char *dir = 0, int force = 1);
    int removeMemPoolFile(const char *dir = 0, int force = 1);
    int removeTransactionManagerFile(const char *dir = 0, int force = 1);
    int removeLoggingFile(const char *dir = 0, int force = 1);

    static DBApp *getApp();
    static int runtimeError(const char *caller, int err, int in_destructor = 0);

private:
    // no copying
    DBApp(const DBApp &);
    operator = (const DBApp &);

    int commonCreateTable(DBTable *table_returned,
                          const char *fname,
                          int type,
                          int mode,
                          DBOpenParams params);

    int commonOpenTable(DBTable *table_returned,
                        const char *fname,
                        int type,
                        DBOpenParams params);

    DEFINE_DB_CLASS(DBApp);

    DBAppOptions::ErrorModel errorModel_;
    char *name_;
};

//
// A symbol table that specifies locations of file system objects
// for use by the application.
//
class _exported DBConfig
{
    friend DBApp;

public:
    DBConfig();
    virtual ~DBConfig();

    // Set the directory for log files for the database.
    //
    void setLogDirectory(const char *dir);
    const char *getLogDirectory() const;

    // Set the directory for temporary files for the database.
    //
    void setTempDirectory(const char *dir);
    const char *getTempDirectory() const;

    // Add a directory to the list of directories used for
    // data files.  The first directory in the list is
    // used for creating new files.  Other directories
    // in the list are used for accessing existing files.
    //
    void setDataDirectoryForCreation(const char *dir);
    void addDataDirectory(const char *dir);
    const char *getDataDirectory(int which = 0) const;

    // General purpose parameter access method.
    //
    void setParam(const char *name, const char *value, int replace = 1);
    const char *getParam(const char *name, int which = 0) const;

private:
    int findParamIndex(const char *name, int which) const;

    // no copying
    DBConfig(const DBConfig &);
    operator = (const DBConfig &);

    int len_;
    char **params_;
};

////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
//
// Table access classes
//

//
// Represents a database table = a set of keys with associated values.
//
class _exported DBTable
{
    friend DBApp;

public:
    DBTable();
    virtual ~DBTable();

    int close(int noSync = 0);

    int get(DBThunk &key,
            DBThunk &value,
            DBTransaction &txn = DBTransaction::NoTransaction);

    int put(DBThunk key,
            DBThunk value,
            DBTransaction &txn = DBTransaction::NoTransaction,
            int noOverwrites = 0);

    int getIterator(DBIterator *iter_returned,
                    DBTransaction &txn = DBTransaction::NoTransaction);

    int remove(DBThunk key,
               DBTransaction &txn = DBTransaction::NoTransaction);

    // get underlying file descriptor
    //
    int getFileDescriptor(int *fd_returned);

    // synchronize inmemory tables to disk
    //
    int sync();

    // gather statistics
    //
    int stat(void *gsp, void *lsp, void *(*alternate_malloc)(size_t) = 0);

    DBTable(const DBTable &);
    DBTable &operator = (const DBTable &);

    virtual int kind() const;         // used internally
    int isValid() const;

private:
    DEFINE_DB_CLASS(DBTable);
};

class _exported DBBTreeTable : public DBTable
{
    friend DBApp;

public:
    DBBTreeTable();
    virtual ~DBBTreeTable();
    DBBTreeTable(const DBBTreeTable &);
    DBBTreeTable &operator = (const DBBTreeTable &);
    virtual int kind() const;         // used internally
};

class _exported DBHashTable : public DBTable
{
    friend DBApp;

public:
    DBHashTable();
    virtual ~DBHashTable();
    DBHashTable(const DBHashTable &);
    DBHashTable &operator = (const DBHashTable &);
    virtual int kind() const;         // used internally
};

class _exported DBRecnoTable : public DBTable
{
    friend DBApp;

public:
    DBRecnoTable();
    virtual ~DBRecnoTable();
    DBRecnoTable(const DBRecnoTable &);
    DBRecnoTable &operator = (const DBRecnoTable &);
    virtual int kind() const;         // used internally
};

//
// A chunk of data, maybe a key or value.
//
class _exported DBThunk
{
    friend DBIterator;
    friend DBTable;

public:
    int hasData() const;

    void setData(void *data, size_t size);
    void setData(void *data);
    void setSize(size_t size);
    void *getData() const;
    size_t getSize() const;

    void setUserBufferSize(size_t);
    size_t getUserBufferSize() const;
    size_t getRecordSize() const;
    size_t getRecordOffset() const;

    void setReturnInAllocatedMemory(int onOrOff);
    int getReturnInAllocatedMemory() const;
    void setReturnInUserBuffer(int onOrOff);
    int getReturnInUserBuffer() const;
    void setPartial(int onOrOff);
    int getPartial() const;

    DBThunk(void *data, size_t size);
    DBThunk();
    virtual ~DBThunk();
    DBThunk(const DBThunk &);
    DBThunk &operator = (const DBThunk &);

private:
    void allocate();
    void deallocate();

    DEFINE_DB_CLASS(DBThunk);
};

class _exported DBIterator
{
    friend DBTable;

public:
    int getKey(DBThunk *key_returned);
    int getValue(DBThunk *value_returned);
    int get(DBThunk *key_returned, DBThunk *value_returned);
    int putCurrent(DBThunk value);
    int putAfter(DBThunk key, DBThunk value);
    int putBefore(DBThunk key, DBThunk value);

    int next();             // returns DB_NOTFOUND if no more, or error code
    int previous();         // returns DB_NOTFOUND if no more, or error code
    int noMore();           // true if next() or prev() have advanced too far

    int close();

    int setToBeginning();
    int setToEnd();

    // By default, the iterator is set to the first element.
    //
    DBIterator();
    virtual ~DBIterator();

    int isValid() const;

private:
    int commonIterate(int direction, const char *name);
    int commonPut(DBThunk key, DBThunk value,
                  int direction, const char *name);

    DEFINE_DB_CLASS(DBIterator);

    DBThunk currentKey_;
    DBThunk currentValue_;
};

class _exported DBSequenceNumber
{
private:
    DEFINE_DB_CLASS(DBSequenceNumber);
};

////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
//
// Library information classes
//

//
// Encapsulates version information for the db library.
//
class _exported DBVersion
{
public:
    DBVersion(int major, int minor, int patch);

    static const char *GetVersionString();
    static DBVersion &GetVersion();

    // Get the major, minor and patch levels for this version.
    //
    int getMajor() const;
    int getMinor() const;
    int getPatch() const;

private:
    // no copying
    DBVersion(const DBVersion &);
    operator = (const DBVersion &);

    int major_;
    int minor_;
    int patch_;
};
#endif /* !_DB_CXX_H_ */
