/*-
 * See the file LICENSE for redistribution information.
 *
 * Copyright (c) 1996, 1997
 *	Sleepycat Software.  All rights reserved.
 */

#include "config.h"

#ifndef lint
static const char sccsid[] = "@(#)db_os_fid.c	10.5 (Sleepycat) 7/2/97";
#endif /* not lint */

#ifndef NO_SYSTEM_INCLUDES
#include <sys/types.h>
#include <sys/stat.h>

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#endif

#include "db_int.h"
#include "db_page.h"
#include "os_ext.h"
#include "common_ext.h"

/*
 * __db_fileid --
 *	Return a unique identifier for a file.
 *
 * PUBLIC: int __db_fileid __P((DB_ENV *, const char *, u_int8_t **, size_t *));
 */
int
__db_fileid(dbenv, fname, fidp, lenp)
	DB_ENV *dbenv;
	const char *fname;
	u_int8_t **fidp;
	size_t *lenp;
{
#ifdef _WIN32
	/*
	 * The documentation for GetFileInformationByHandle() states that the
	 * inode-type numbers are not constant between processes.  Actually,
	 * they are, they're the NTFS MFT indexes.  So, this works on NTFS,
	 * but perhaps not on other platforms, and perhaps not over a network.
	 * Can't think of a better solution right now.
	 */
	u_int8_t *bp = 0;
	int fd = 0;
	HANDLE fh = 0;
	BY_HANDLE_FILE_INFORMATION fi;
	BOOL retval = FALSE;

	/* first we open the file, because we're not given a handle to it */
	fd = open(fname,_O_RDONLY,_S_IREAD);
	if (-1 == fd) {
		/* If we can't open it, we're in trouble */
		return (errno);
	}

	/* File open, get its info */
	fh = (HANDLE)_get_osfhandle(fd);
	if ((HANDLE)(-1) != fh) {
		retval = GetFileInformationByHandle(fh,&fi);
	}
	close(fd);

	/*
	 * We want the three 32-bit words which tell us the volume ID and
	 * the file ID.  We make a crude attempt to copy the bytes over to
	 * the callers buffer.
	 *
	 * DBDB: really we should ensure that the bytes get packed the same
	 * way on all compilers, platforms etc.
	 */
	if ( ((HANDLE)(-1) != fh) && (TRUE == retval) ) {
		bp = malloc(3 * sizeof (DWORD) );
		if (0 != bp) {
			*((u_int32_t*)(bp)) = fi.nFileIndexLow;
			*((u_int32_t*)(bp+sizeof(DWORD))) = fi.nFileIndexHigh;
			*((u_int32_t*)(bp+(2*sizeof(DWORD)))) =
			    fi.dwVolumeSerialNumber;
			*lenp = 3 * sizeof (DWORD);
		  }
	}

	*fidp = bp;
	return (0);
#else
	struct stat sb;
	int i;
	u_int8_t *bp, *p, *t;

	/*
	 * !!!
	 * On UNIX, we use a dev/inode pair.
	 *
	 * XXX
	 * Note, this isn't perfectly safe, as it's possible to have dev/ino
	 * pairs that aren't unique when the files reside on NFS filesystems.
	 * It's the best we can do portably, however.
	 */
	if (stat(fname, &sb)) {
		__db_err(dbenv, "%s: %s", fname, strerror(errno));
		return (errno);
	}
	if ((bp = (u_int8_t *)
	    malloc(sizeof(sb.st_ino) + sizeof(sb.st_dev))) == NULL) {
		__db_err(dbenv, "%s", strerror(ENOMEM));
		return (ENOMEM);
	}

	/*
	 * Use the inode first and in reverse order, hopefully putting the
	 * distinguishing information early in the string.
	 */
	t = bp;
	for (p = (u_int8_t *)&sb.st_ino +
	    sizeof(sb.st_ino), i = 0; i < sizeof(sb.st_ino); ++i)
		*t++ = *--p;
	for (p = (u_int8_t *)&sb.st_dev +
	    sizeof(sb.st_dev), i = 0; i < sizeof(sb.st_dev); ++i)
		*t++ = *--p;
	*lenp = sizeof(sb.st_ino) + sizeof(sb.st_dev);

	*fidp = bp;
	return (0);
#endif
}
